@@ -9,7 +9,9 @@
 
 # %% Libraries
 import wget
+import sys
 import os
+from tqdm import tqdm
 import numpy as np
 import pandas as pd
 from sklearn.preprocessing import LabelEncoder
@@ -20,7 +22,7 @@ onehot_encoder = OneHotEncoder(sparse=False, categories='auto')
 
 
 # %% Dataframe to one-hot
-def df2onehot(df, dtypes='pandas', y_min=None, perc_min_num=None, hot_only=True, list_expand=True, excl_background=None, verbose=3):
+def df2onehot(df, dtypes='pandas', y_min=None, perc_min_num=None, hot_only=True, deep_extract=False, excl_background=None, verbose=3):
     """Convert dataframe to one-hot matrix.
 
     Parameters
@@ -35,8 +37,9 @@ def df2onehot(df, dtypes='pandas', y_min=None, perc_min_num=None, hot_only=True,
         Force column (int or float) to be numerical if unique non-zero values are above percentage. The default is None. Alternative can be 0.8
     hot_only : bool [True, False], optional
         When True; the output of the onehot matrix exclusively contains categorical values that are transformed to one-hot. The default is True.
-    list_expand : bool [True, False], optional
-        Expanding of columns that contain lists of strings.. The default is True.
+    deep_extract : bool [False, True] (default : False)
+        True: Extract information from a vector that contains a list/array/dict.
+        False: converted to a string and treated as catagorical ['cat'].
     excl_background : list or None, [0], [0, '0.0', 'male', ...], optional
         Remove values/strings that labeled as background. As an example, in a two-class approach with [0,1], the 0 is usually the background and not of interest. The default is None.
     verbose : int, optional
@@ -66,31 +69,32 @@ def df2onehot(df, dtypes='pandas', y_min=None, perc_min_num=None, hot_only=True,
     args['dtypes'] = dtypes
     args['verbose'] = verbose
     args['perc_min_num'] = perc_min_num
-    args['deep_extract'] = list_expand
+    args['deep_extract'] = deep_extract
     args['excl_background'] = excl_background
     labx = []
+    disable = (True if (verbose==0 or verbose>3) else False)
 
     # Reset index
     df.reset_index(drop=True, inplace=True)
-    # Remove non-ascii chars
-    # df = _remove_non_ascii(df)
     # Determine Dtypes
-    df, dtypes = set_dtypes(df, args['dtypes'], is_list=args['deep_extract'], perc_min_num=args['perc_min_num'], verbose=args['verbose'])
+    df, dtypes = set_dtypes(df, args['dtypes'], deep_extract=args['deep_extract'], perc_min_num=args['perc_min_num'], verbose=args['verbose'])
     # If any column is a list, also expand the list!
-    df, dtypes = _deep_extract(df, dtypes, perc_min_num=args['perc_min_num'], verbose=args['verbose'])
+    if args['deep_extract']:
+        df, dtypes = _deep_extract(df, dtypes, perc_min_num=args['perc_min_num'], verbose=args['verbose'])
 
     # Make empty frames
+    maxstring=50
     out_numeric = pd.DataFrame()
     out_onehot = pd.DataFrame()
-    max_str_len = np.max(list(map(len, df.columns.values.astype(str).tolist())))+2
+    max_str_len = np.minimum(np.max(list(map(len, df.columns.values.astype(str).tolist())))+2, maxstring)
 
     # Run over all columns
-    for i in np.arange(0, df.shape[1]):
-        makespaces = ''.join(['.'] * (max_str_len - len(df.columns[i])))
+    for i in tqdm(np.arange(0, df.shape[1]), disable=disable):
+        makespaces = ''.join(['.'] * np.minimum( (max_str_len - len(df.columns[i])), maxstring) )
         # Do not touch a float
         if 'float' in str(df.dtypes[i]):
             # if verbose>=3: print('[df2onehot] >Working on %s' %(df.columns[i]))
-            if verbose>=3: print('[df2onehot] >Working on %s%s[float]' %(df.columns[i], makespaces))
+            if verbose>=4: print('[df2onehot] >Processing: %s%s [float]' %(df.columns[i][0:maxstring], makespaces))
             out_numeric[df.columns[i]] = df.iloc[:,i]
             if hot_only is False:
                 out_onehot[df.columns[i]] = df.iloc[:,i]
@@ -102,7 +106,7 @@ def df2onehot(df, dtypes='pandas', y_min=None, perc_min_num=None, hot_only=True,
             # integer_encoded = set_y(integer_encoded, y_min=y_min, numeric=True, verbose=0)
             out_numeric[df.columns[i]] = integer_encoded
             out_numeric[df.columns[i]] = out_numeric[df.columns[i]].astype('category')
-            if verbose>=3: print('[df2onehot] >Working on %s%s[%.0f]' %(df.columns[i], makespaces, len(np.unique(integer_encoded)) ))
+            if verbose>=4: print('[df2onehot] >Processing: %s%s [%.0f]' %(df.columns[i][0:maxstring], makespaces, len(np.unique(integer_encoded)) ))
 
             # Contains a single value or is bool
             if (len(np.unique(integer_encoded))<=1) or (str(df.dtypes[i])=='bool'):
@@ -130,7 +134,7 @@ def df2onehot(df, dtypes='pandas', y_min=None, perc_min_num=None, hot_only=True,
                     out_numeric[df.columns[i]] = (onehot_encoded * np.arange(1,onehot_encoded.shape[1] + 1)).sum(axis=1)
 
     uiy, ycounts = np.unique(labx, return_counts=True)
-    if verbose >=3: print('[df2onehot] >\n[df2onehot] >Total onehot features: %.0d' %(np.sum(ycounts)))
+    if verbose >=3: print('[df2onehot] >Total onehot features: %.0d' %(np.sum(ycounts)))
     # idx = np.argsort(ycounts)[::-1]
     # print(np.c_[uiy[idx], ycounts[idx]])
 
@@ -151,32 +155,42 @@ def df2onehot(df, dtypes='pandas', y_min=None, perc_min_num=None, hot_only=True,
 
 
 # %% Convert str/float/int to type
-def _col2type(df, dtypes, idx):
-    gettype = dtypes[idx]
+def _col2type(dfc, verbose=3):
+    # dtype = dtypes[idx]
     # Gather only the not NaN rows
-    Inan = df.iloc[:, idx].isna()
+    Inan = dfc.isna()
     # Grap only unique elements from the combined set of lists
-    dfcol = df.iloc[~Inan.values, idx].copy()
+    dfcol = dfc.iloc[~Inan.values].copy()
+    # Remvoe empty lists
+    # Inan = dfcol.apply(len)==0
+    # dfcol = dfcol[~Inan.values]
+    
     # If any str, float or int elements is fount, convert to list
-    if gettype=='list':
-        typing = list(map(type, dfcol.values))
-        Iloc = np.logical_or(list(map(lambda x: isinstance(str(), x), typing)), np.logical_or(list(map(lambda x: isinstance(float(), x), typing)), list(map(lambda x: isinstance(int(), x), typing))))
-        if np.any(Iloc):
-            # Convert to string! This is required because the colnames also converts all to string.
-            dfcol.loc[Iloc] = list(map(lambda x: [str(x)], dfcol.loc[Iloc]))
-            df.iloc[dfcol.loc[Iloc].index,idx] = dfcol.loc[Iloc]
+    # if dtype=='list':
+    typing = list(map(type, dfcol.values))
+    Iloc = np.logical_or(list(map(lambda x: isinstance(str(), x), typing)), np.logical_or(list(map(lambda x: isinstance(float(), x), typing)), list(map(lambda x: isinstance(int(), x), typing))))
+    if np.any(Iloc):
+        # Convert to string! This is required because the colnames also converts all to string.
+        dfcol.loc[Iloc] = list(map(lambda x: [str(x)], dfcol.loc[Iloc]))
+        dfc.iloc[dfcol.loc[Iloc].index] = dfcol.loc[Iloc]
 
     # Get all unique elements
-    uifeat = _get_unique_elements(dfcol)
+    uifeat = _get_unique_elements(dfcol, verbose=verbose)
     # Return
-    return(df, uifeat)
+    return(dfc, uifeat)
+
+def _get_unique_elements(dfcol, verbose=3):
+    try:
+        # Get all unique elements
+        listvector = list(map(lambda x: list(x) , dfcol))
+        listvector = sum(listvector, [])
+        # All elements are converted to string!
+        uifeat = np.unique(listvector)
+    except:
+        # if verbose>=1: print('[df2onehot] >Error catched:' %(str(sys.exc_info()[0])))
+        if verbose>=1: print('[df2onehot] >Error catched.')
+        uifeat = None
 
-def _get_unique_elements(dfcol):
-    # Get all unique elements
-    listvector = list(map(lambda x: list(x) , dfcol))
-    listvector = sum(listvector, [])
-    # All elements are converted to string!
-    uifeat = np.unique(listvector)
     return uifeat
 
 def _array2df(df, uifeat, idx):
@@ -239,56 +253,80 @@ def dict2df(dfc):
 
 # %%
 def _deep_extract(df, dtypes, perc_min_num=None, verbose=3):
-    # Check for any lists in dtypes
-    Ilist = np.isin(dtypes,'list')
-    Idict = np.isin(dtypes,'dict')
-    dftot1 = pd.DataFrame()
-    dftot2 = pd.DataFrame()
-    idxrem1 = []
-    idxrem2 = []
+    # Extract dict
+    dftot1, idxrem1 = _extract_dict(df, dtypes, verbose=verbose)
+    # Extract lists
+    dftot2, idxrem2 = _extract_list(df, dtypes, verbose=verbose)
+    # Combine the extracts
+    df, dtypes = _extract_combine(df, dtypes, dftot1, dftot2, idxrem1, idxrem2, perc_min_num, verbose=verbose)
 
+    # Return
+    if df.shape[1]!=len(dtypes): raise Exception('[df2onehot] >Error: size of dtypes and dataframe does not match.')
+    return(df, dtypes)
+
+# %%
+def _extract_dict(df, dtypes, verbose=3):
+    disable = (True if (verbose==0 or verbose>3) else False)
+    dfout = pd.DataFrame()
+    idxrem = []
+    Idict = np.isin(dtypes,'dict')
 
     # Expand dict
     if np.any(Idict):
-        if verbose >=3: print('[df2onehot] >\n[df2onehot] >Starting deep extraction..')
+        if verbose >=3: print('[df2onehot] >Deep extract..')
         idxCol = np.where(Idict)[0]
         max_str_len = np.max(list(map(len, df.columns[idxCol].values.astype(str).tolist())))
         # Expand every columns that contains dict
-        for idx in idxCol:
+        for idx in tqdm(idxCol, disable=disable):
             makespaces = ''.join([' '] * (max_str_len - len(df.columns[idx])))
             try:
                 dfc, idxempty = dict2df(df.iloc[:, idx])
                 # dfc = pd.DataFrame.from_records(df.iloc[:,idx])
-                if verbose>=3: print('[df2onehot] >[%s]%s >deep extract > [%s]  [%d]' %(df.columns[idx], makespaces, dtypes[idx], dfc.shape[1]))
+                if verbose>=4: print('[df2onehot] >[%s]%s >deep extract > [%s]  [%d]' %(df.columns[idx], makespaces, dtypes[idx], dfc.shape[1]))
             except:
-                if verbose>=3: print('[df2onehot] >[%s]%s >deep extract > [failed]' %(df.columns[idx], makespaces))
+                if verbose>=4: print('[df2onehot] >[%s]%s >deep extract > [failed]' %(df.columns[idx], makespaces))
                 # dfc = df.iloc[:,idx].astype(str)
                 # dfc = dfc.apply(_remove_non_ascii)
 
-            # Remove non ascii chars
-            # dfc = _remove_non_ascii(dfc)
             # Combine extracted columns into big dataframe
-            dftot1 = pd.concat([dftot1, dfc], axis=1)
+            dfout = pd.concat([dfout, dfc], axis=1)
             # Add idx to remove
-            idxrem1.append(idx)
+            idxrem.append(idx)
+    
+    return dfout, idxrem
+
+# %%
+def _extract_list(df, dtypes, verbose=3):
+    disable = (True if (verbose==0 or verbose>3) else False)
+    Ilist = np.isin(dtypes,'list')
+    dfout = pd.DataFrame()
+    idxrem = []
 
     # Expand list
     if np.any(Ilist):
         idxCol = np.where(Ilist)[0]
         max_str_len = np.max(list(map(len, df.columns[idxCol].values.astype(str).tolist())))
         # Expand every columns that contains either list
-        for idx in idxCol:
+        for idx in tqdm(idxCol, disable=disable):
             makespaces = ''.join([' '] * (max_str_len - len(df.columns[idx])))
-            # Convert str/float/int to type
-            df, uifeat = _col2type(df, dtypes, idx)
+            # Convert str/float/int to list
+            # df, uifeat = _col2type(df, dtypes, idx)
+            df.iloc[:,idx], uifeat = _col2type(df.iloc[:,idx], verbose=verbose)
+            
             # Convert column into onehot
-            dfc = _array2df(df, uifeat, idx)
-            # Combine hot-dataframes into one big dataframe
-            dftot2 = _concat(dftot2, dfc)
-            # Add idx to remove
-            idxrem2.append(idx)
-            if verbose>=3: print('[df2onehot] >[%s]%s >deep extract > [%s]  [%d]' %(df.columns[idx], makespaces, dtypes[idx], dfc.shape[1]))
+            if uifeat is not None:
+                dfc = _array2df(df, uifeat, idx)
+                # Combine hot-dataframes into one big dataframe
+                dfout = _concat(dfout, dfc)
 
+            # Add idx to remove
+            idxrem.append(idx)
+            if verbose>=4: print('[df2onehot] >[%s]%s >deep extract > [%s]  [%d]' %(df.columns[idx], makespaces, dtypes[idx], dfc.shape[1]))
+    
+    return dfout, idxrem
+
+# %%
+def _extract_combine(df, dtypes, dftot1, dftot2, idxrem1, idxrem2, perc_min_num, verbose=3):
     # Drop columns that are expanded
     idxrem = idxrem1+idxrem2
     if len(idxrem)>0:
@@ -302,22 +340,23 @@ def _deep_extract(df, dtypes, perc_min_num=None, verbose=3):
         # Remove repetative column names
         dftot = _make_columns_unique(dftot, verbose=verbose)
         # Set dtypes
-        dftot, dtypest = set_dtypes(dftot, perc_min_num=perc_min_num, verbose=3)
+        dftot, dtypest = set_dtypes(dftot, perc_min_num=perc_min_num, deep_extract=False, verbose=3)
         # Combine into dataframe
         df = pd.concat([df, dftot], axis=1)
         dtypes = dtypes + dtypest
-        if verbose>=3: print('[df2onehot] >[%d] additional columns extracted by deep extract.' %(dftot1.shape[1]+dftot2.shape[1]))
-
-    # Return
-    if df.shape[1]!=len(dtypes): raise Exception('[df2onehot] >Error: size of dtypes and dataframe does not match.')
-    return(df, dtypes)
-
+        if verbose>=3: print('[df2onehot] >Deep extract extracted: [%d] features.' %(dftot1.shape[1]+dftot2.shape[1]))
+    return df, dtypes
 
 # %% Remove repetative column
 def _make_columns_unique(dftot, verbose=3):
     columns = dftot.columns.value_counts()
     columns = columns[columns.values>1].index.values
     if verbose>=3: print('[df2onehot] >[%d] repetative columns detected and a single one is taken: %s' %(len(columns), columns))
+
+    # for column in columns:
+    #     dfc = dftot[column]
+    #     dfmerged = dfc.stack().groupby(level=0).apply(lambda x: x.unique().tolist())
+
     _, uiidx = np.unique(dftot.columns, return_index=True)
     dftot = dftot.iloc[:,np.sort(uiidx)]
     return dftot
